@@ -1,6 +1,6 @@
 {- git-annex command-line option parsing
  -
- - Copyright 2010-2023 Joey Hess <id@joeyh.name>
+ - Copyright 2010-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -40,6 +40,7 @@ import qualified Types.Backend as Backend
 import Utility.HumanTime
 import Utility.DataUnits
 import Annex.Concurrent
+import Remote.List
 
 -- Options that are accepted by all git-annex sub-commands,
 -- although not always used.
@@ -55,6 +56,11 @@ gitAnnexCommonOptions = commonOptions ++
 		<> help "override minimum number of copies"
 		<> hidden
 		)
+	, annexFlag (setrebalance True)
+		( long "rebalance" 
+		<> help "move content as needed to improve balance"
+		<> hidden
+		)
 	, annexOption (setAnnexState . Remote.forceTrust Trusted) $ strOption
 		( long "trust" <> metavar paramRemote
 		<> help "deprecated, does not override trust setting"
@@ -102,15 +108,19 @@ gitAnnexCommonOptions = commonOptions ++
   where
 	setnumcopies n = setAnnexRead $ \rd -> rd { Annex.forcenumcopies = Just $ configuredNumCopies n }
 	setmincopies n = setAnnexRead $ \rd -> rd { Annex.forcemincopies = Just $ configuredMinCopies n }
+	setrebalance v = setAnnexRead $ \rd -> rd { Annex.rebalance = v }
 	setuseragent v = setAnnexRead $ \rd -> rd { Annex.useragent = Just v }
 	setdesktopnotify v = setAnnexRead $ \rd -> rd { Annex.desktopnotify = Annex.desktopnotify rd <> v }
 	setgitconfig v = Annex.addGitConfigOverride v
 
 {- Parser that accepts all non-option params. -}
 cmdParams :: CmdParamsDesc -> Parser CmdParams
-cmdParams paramdesc = many $ argument str
+cmdParams paramdesc = cmdParamsWithCompleter paramdesc completeFiles
+
+cmdParamsWithCompleter :: String -> Mod ArgumentFields String -> Parser CmdParams
+cmdParamsWithCompleter paramdesc completers = many $ argument str
 	( metavar paramdesc
-	<> action "file"
+	<> completers
 	)
 
 parseAutoOption :: Parser Bool
@@ -162,40 +172,55 @@ parseToOption = strOption
 	<> completeRemotes
 	)
 
+parseFromAnywhereOption :: Parser Bool
+parseFromAnywhereOption = switch
+	( long "from-anywhere"
+	<> help "from any remote"
+	)
+
 parseRemoteOption :: Parser RemoteName
 parseRemoteOption = strOption
 	( long "remote" <> metavar paramRemote
 	<> completeRemotes
 	)
 
--- | From or to a remote, or both, or a special --to=here
+-- | --from or --to a remote, or both, or a special --to=here,
+-- or --from-anywhere --to remote.
 data FromToHereOptions 
 	= FromOrToRemote FromToOptions
 	| ToHere
 	| FromRemoteToRemote (DeferredParse Remote) (DeferredParse Remote)
+	| FromAnywhereToRemote (DeferredParse Remote)
 
 parseFromToHereOptions :: Parser (Maybe FromToHereOptions)
 parseFromToHereOptions = go
 	<$> optional parseFromOption
 	<*> optional parseToOption
+	<*> parseFromAnywhereOption
   where
-	go (Just from) (Just to) = Just $ FromRemoteToRemote
+	go _ (Just to) True = Just $ FromAnywhereToRemote
+		(mkParseRemoteOption to)
+	go (Just from) (Just to) _ = Just $ FromRemoteToRemote
 		(mkParseRemoteOption from)
 		(mkParseRemoteOption to)
-	go (Just from) Nothing = Just $ FromOrToRemote
+	go (Just from) Nothing _ = Just $ FromOrToRemote
 		(FromRemote $ mkParseRemoteOption from)
-	go Nothing (Just to) = Just $ case to of
+	go Nothing (Just to) _ = Just $ case to of
 		"here" -> ToHere
 		"." -> ToHere
 		_ -> FromOrToRemote $ ToRemote $ mkParseRemoteOption to
-	go Nothing Nothing = Nothing
+	go Nothing Nothing _ = Nothing
 
 instance DeferredParseClass FromToHereOptions where
-	finishParse (FromOrToRemote v) = FromOrToRemote <$> finishParse v
+	finishParse (FromOrToRemote v) = 
+		FromOrToRemote <$> finishParse v
 	finishParse ToHere = pure ToHere
-	finishParse (FromRemoteToRemote v1 v2) = FromRemoteToRemote
-		<$> finishParse v1
-		<*> finishParse v2
+	finishParse (FromRemoteToRemote v1 v2) = 
+		FromRemoteToRemote
+			<$> finishParse v1
+			<*> finishParse v2
+	finishParse (FromAnywhereToRemote v) = 
+		FromAnywhereToRemote <$> finishParse v
 
 -- Options for acting on keys, rather than work tree files.
 data KeyOptions
@@ -266,6 +291,7 @@ annexedMatchingOptions = concat
 keyMatchingOptions :: [AnnexOption]
 keyMatchingOptions = concat
 	[ keyMatchingOptions'
+	, sizeMatchingOptions Limit.LimitAnnexFiles
 	, anythingNothingOptions
 	, combiningOptions 
 	, timeLimitOption 
@@ -277,12 +303,12 @@ keyMatchingOptions' :: [AnnexOption]
 keyMatchingOptions' = 
 	[ annexOption (setAnnexState . Limit.addIn) $ strOption
 		( long "in" <> short 'i' <> metavar paramRemote
-		<> help "match files present in a remote"
+		<> help "match files present in a repository"
 		<> hidden
 		<> completeRemotes
 		)
 	, annexOption (setAnnexState . Limit.addCopies) $ strOption
-		( long "copies" <> short 'C' <> metavar paramRemote
+		( long "copies" <> short 'C' <> metavar paramNumber
 		<> help "skip files with fewer copies"
 		<> hidden
 		)
@@ -322,6 +348,11 @@ keyMatchingOptions' =
 		<> help "match files with attached metadata"
 		<> hidden
 		)
+	, annexOption (setAnnexState . Limit.addUrl) $ strOption
+		( long "url" <> metavar paramGlob
+		<> help "match files by url"
+		<> hidden
+		)
 	, annexFlag (setAnnexState Limit.Wanted.addWantGet)
 		( long "want-get"
 		<> help "match files the local repository wants to get"
@@ -350,6 +381,21 @@ keyMatchingOptions' =
 			<> help "match files accessed within a time interval"
 			<> hidden
 			)
+	, annexOption (setAnnexState . Limit.addPresentSince) $ strOption
+		( long "presentsince" <> metavar paramValue
+		<> help "matches files present in a repository throughout a time interval"
+		<> hidden
+		)
+	, annexOption (setAnnexState . Limit.addLackingSince) $ strOption
+		( long "lackingsince" <> metavar paramValue
+		<> help "matches files not present in a repository throughout a time interval"
+		<> hidden
+		)
+	, annexOption (setAnnexState . Limit.addChangedSince) $ strOption
+		( long "changedsince" <> metavar paramValue
+		<> help "matches files whose presence changed during a time interval"
+		<> hidden
+		)
 	, annexOption (setAnnexState . Limit.addMimeType) $ strOption
 		( long "mimetype" <> metavar paramGlob
 		<> help "match files by mime type"
@@ -370,6 +416,11 @@ keyMatchingOptions' =
 		<> help "match files that are locked"
 		<> hidden
 		)
+	, annexFlag (setAnnexState Limit.addExpectedPresent)
+		( long "expected-present"
+		<> help "match files expected to be present"
+		<> hidden
+		)
 	]
 
 -- Options to match files which may not yet be annexed.
@@ -398,7 +449,11 @@ fileMatchingOptions' lb =
 		<> help "limit to files whose content is the same as another file matching the glob pattern"
 		<> hidden
 		)
-	, annexOption (setAnnexState . Limit.addLargerThan lb) $ strOption
+	] ++ sizeMatchingOptions lb
+
+sizeMatchingOptions :: Limit.LimitBy -> [AnnexOption]
+sizeMatchingOptions lb =
+	[ annexOption (setAnnexState . Limit.addLargerThan lb) $ strOption
 		( long "largerthan" <> metavar paramSize
 		<> help "match files larger than a size"
 		<> hidden
@@ -476,15 +531,19 @@ jsonProgressOption =
 -- action in `allowConcurrentOutput`.
 jobsOption :: [AnnexOption]
 jobsOption = 
-	[ annexOption (setAnnexState . setConcurrency . ConcurrencyCmdLine) $ 
-		option (maybeReader parseConcurrency)
-			( long "jobs" <> short 'J' 
-			<> metavar (paramNumber `paramOr` "cpus")
-			<> help "enable concurrent jobs"
-			<> hidden
-			)
+	[ annexOption (setAnnexState . setConcurrency . ConcurrencyCmdLine)
+		jobsOptionParser
 	]
 
+jobsOptionParser :: Parser Concurrency
+jobsOptionParser = 
+	option (maybeReader parseConcurrency)
+		( long "jobs" <> short 'J' 
+		<> metavar (paramNumber `paramOr` "cpus")
+		<> help "enable concurrent jobs"
+		<> hidden
+		)
+
 timeLimitOption :: [AnnexOption]
 timeLimitOption = 
 	[ annexOption settimelimit $ option (eitherReader parseDuration)
@@ -544,14 +603,34 @@ parseDaemonOptions canstop
 		)
 
 completeRemotes :: HasCompleter f => Mod f a
-completeRemotes = completer $ mkCompleter $ \input -> do
-	r <- maybe (pure Nothing) (Just <$$> Git.Config.read)
-		=<< Git.Construct.fromCwd
-	return $ filter (input `isPrefixOf`) $
-		mapMaybe remoteKeyToRemoteName $
-			filter isRemoteUrlKey $
-				maybe [] (M.keys . config) r
-		
+completeRemotes = completer $ mkCompleter $ \input ->
+	Git.Construct.fromCwd >>= \case
+		Nothing -> return []
+		Just g -> completeRemotes' g input
+
+completeRemotes' :: Repo -> [Char] -> IO [[Char]]
+completeRemotes' g input = do
+	g' <- Git.Config.read g
+	state <- Annex.new g'
+	Annex.eval state $ do
+		Annex.setOutput QuietOutput
+		gc <- Annex.getGitConfig
+		if isinitialized gc
+			then do
+				rs <- remoteList
+				matches $ map Remote.name rs
+			else matches $
+				mapMaybe remoteKeyToRemoteName $
+					filter isRemoteUrlKey $ 
+						M.keys $ config g
+  where
+	isinitialized gc = annexUUID gc /= NoUUID && isJust (annexVersion gc)
+	matches = return . filter (input `isPrefixOf`)
+
 completeBackends :: HasCompleter f => Mod f a
 completeBackends = completeWith $
 	map (decodeBS . formatKeyVariety . Backend.backendVariety) Backend.builtinList
+
+completeFiles :: HasCompleter f => Mod f a
+completeFiles = action "file"
+
